"""
main submodule, implementing / handling the SI Station
"""

import time
from datetime import datetime, timedelta
from typing import List, Optional, Union

from serial import Serial, SerialException

from sipyconfig.card import SI5, SI6, SI8To10, SICard
from sipyconfig.comms import ACK, NOACK, Command, receive_command, send_command
from sipyconfig.enums import (CPC, SETTINGS, SiCardInserted, SiCardRemoved,
                              SiError, com, memaddr, memlen, misc, mode_rev,
                              proto, stationmode, SiCardAutoSend, siacmode,
                              siacmode_rev, MODELID, model_rev)
from sipyconfig.utils import (apply_bit_mask, array_or, bytes_to_int, is_siac_mode,
                              is_siac_special_mode, log2)

STANDARD_PORT = "/dev/ttyUSB{0:d}"
STANDARD_SERIAL_TIMEOUT = 5
BAUDRATE_EXTENDED = 38400
BAUDRATE_STANDARD = 4800
STANDARD_MAX_COMMAND_TRIES = 32

class SiUSBStation():
    """main class for interfacing with the SI Station
    NOTE: only stations with extended protocol are currently supported"""
    _dev: 'Serial' = None
    _device_port: str
    _extended_protocol: bool = True
    _direct_mode: Optional[bool] = None

    _model_id: MODELID = MODELID.UNKNOWN
    _model_id_str: str = "UNKNOWN"

    _mem_size: int = -1
    _sys_data: List[int]

    _cpc: int = -1
    _cpc_ext_protocol: bool
    _auto_send: bool
    _handshake: bool
    _access_password: bool
    _read_after_punch: bool

    _setting: int = -1
    _optical_feedback: bool
    _acoustic_feedback: bool
    _control_number_high_bit: bool

    _operating_time: int = -1

    _control_number: int = -1
    _mode: stationmode = stationmode.UNKNOWN
    _mode_str: str = "UNKNOWN"
    _siac_mode: siacmode = siacmode.NO_SIAC
    _siac_mode_str: str = "NO_SIAC"

    _si_card: Optional[SICard] = None
    _time: 'datetime'
    _time_read_at: 'datetime'

    _command_tries: int = 0
    _serial_timeout: float = STANDARD_SERIAL_TIMEOUT
    _max_command_tries: int = STANDARD_MAX_COMMAND_TRIES
    _currently_waiting: bool = False

    def __init__(self, device: str = None, mode_direct: bool = True) -> None:
        self._sys_data = []

        self._init_serial(device)
        self._station_detect_procedure()
        self.set_mode_direct(mode_direct)
        self.get_system_info()

    def _init_serial(self, device: str = None, baudrate: int = BAUDRATE_EXTENDED) -> None:
        if self._dev:
            self._dev.close()
        if device:
            self._dev = Serial(device, baudrate=baudrate, timeout=self._serial_timeout)
            if self._dev is None:
                raise SiError("Could not open device!")
            self._device_port = device
            self.flush()
            return
        for i in range(128):
            try:
                self._dev = Serial(STANDARD_PORT.format(i), baudrate, timeout=self._serial_timeout)
                self._device_port = STANDARD_PORT.format(i)
                self.flush()
                return
            except SerialException:
                continue
        raise SiError("Could not open device!")

    def _station_detect_procedure(self) -> bool:
        self._extended_protocol = True
        if self.set_mode_direct():
            return True
        self._init_serial(self._device_port, BAUDRATE_STANDARD)
        if self.set_mode_direct():
            self.set_baudrate()
            self._init_serial(self._device_port, BAUDRATE_EXTENDED)
            self.set_mode_direct()
            return True
        self._extended_protocol = False
        self.set_mode_direct()
        if isinstance(self.receive_command(), NOACK):
            raise ConnectionError("Could not connect to SI-Station!")
        return True

    def set_mode_direct(self, set_mode_direct: bool = True) -> bool:
        """set the station into direct (True) or remote (False) mode"""
        command = com.SET_MS_MODE if self.support_extended_protocol else com.SET_MS_MODE_BASIC
        mode = misc.MODE_DIRECT if set_mode_direct else misc.MODE_REMOTE
        ret = self.send_command(command, [mode], False)
        if isinstance(ret, NOACK):
            self._direct_mode = None
            return False
        if isinstance(ret, Command):
            if ret.is_extended_command:
                self._direct_mode = ret.data[2] == misc.MODE_DIRECT
                self._control_number = (ret.data[0] << 8) + ret.data[1]
            else:
                self._direct_mode = ret.data[1] == misc.MODE_DIRECT
                self._control_number = ret.data[0]
        return True

    def set_baudrate(self, high_baudrate: bool = True) -> None:
        """set the baudrate of the station to the defined lower (False) or higher (True) baudrate"""
        self.send_command(com.SET_BAUDRATE, [int(high_baudrate)], True)

    def get_system_info(self) -> None:
        """retreive the system info from the station"""
        ret = self.send_command(com.GET_SYS_DATA, [memaddr.MEM_SIZE, memlen.MEM_SIZE], True)
        if isinstance(ret, Command):
            self._mem_size = ret.data[misc.DATA_COMMAND_OFFSET]
            ret = self.send_command(com.GET_SYS_DATA, [0x00, self._mem_size], True)
            if isinstance(ret, Command):
                self._sys_data = [x for x in ret.data[misc.DATA_COMMAND_OFFSET:]]

                cn1, cn0 = ret.data[0], ret.data[1]
                self._control_number = (cn1 << 8) + cn0

                self._model_id = MODELID(bytes(self._sys_data[memaddr.MODEL_ID:memaddr.MODEL_ID+memlen.MODEL_ID]))
                self._model_id_str = model_rev.get(self._model_id.value, "UNKNOWN")

                self._mode = stationmode(self._sys_data[memaddr.MODE] & 0x3f)
                if self._mode == stationmode.SIAC_SPECIAL:
                    self._control_number = self._sys_data[memaddr.CONTROL_NUMBER]
                    self._mode_str = mode_rev.get(self._control_number, "UNKNOWN")
                else:
                    self._mode_str = mode_rev.get(self._mode, "UNKNOWN")

                self._siac_mode = siacmode(self._sys_data[memaddr.MODE] & 0xf0)
                self._siac_mode_str = siacmode_rev.get(self._siac_mode, "NO_SIAC")

                self._setting = self._sys_data[memaddr.SETTINGS_FLAG]
                self._optical_feedback = bool(self._setting & SETTINGS.OPTICAL_FEEDBACK)
                self._acoustic_feedback = bool(self._setting & SETTINGS.ACOUSTIC_FEEDBACK)
                self._control_number_high_bit = bool(self._setting & SETTINGS.NUMBER_HIGH_BIT)

                self._cpc = self._sys_data[memaddr.PROTOCOL_MODE_FLAG]
                self._cpc_ext_protocol = bool(self._cpc & CPC.EXTENDED_PROTCOL)
                self._auto_send = bool(self._cpc & CPC.AUTO_SEND)
                self._handshake = bool(self._cpc & CPC.HANDSHAKE)
                self._access_password = bool(self._cpc & CPC.ACCESS_PASSWORD)
                self._read_after_punch = bool(self._cpc & CPC.READ_AFTER_PUNCH)

                self._operating_time = bytes_to_int(self._sys_data[memaddr.ACTIVE_TIME:memaddr.ACTIVE_TIME+memlen.ACTIVE_TIME])
        self.get_time()

    def get_time(self) -> None:
        """get the current time of the station"""
        ret = self.send_command(com.GET_TIME, [], True)
        if isinstance(ret, Command):
            self._time = self._decode_time(ret.data[2:])
            self._time_read_at = datetime.now()

    def _decode_time(self, raw_data: bytes) -> 'datetime':
        year = 2000 + raw_data[0]
        month = raw_data[1]
        day = raw_data[2]
        month = month if 1 <= month <= 12 else 1
        day = day if 1 <= day <= 31 else 1
        seconds = bytes_to_int(raw_data[4:6])
        hour = seconds // 3600
        seconds -= hour * 3600
        minute = seconds // 60
        seconds -= minute * 60
        microseconds = int(raw_data[6] * (1e6 / 256))

        ret_time = datetime(year, month, day, hour, minute, seconds, microseconds)

        if (raw_data[3] << 7) & 0xFF:
            ret_time += timedelta(hours=12)
        return ret_time

    def set_time(self, time_: 'datetime' = None) -> None:
        """set the current time of the station, if time_ is omitted the current computer time will be set"""
        if time_ is None:
            time_ = datetime.now()
        self.send_command(com.SET_TIME, self._encode_time(time_), True)

    def _encode_time(self, time_: 'datetime') -> 'bytes':
        data = [
            time_.year - 2000,
            time_.month,
            time_.day,
            ((time_.weekday() + 1) % 7) << 1
        ]
        if time_.hour >= 12:
            data[3] |= 0x01
            time_ -= timedelta(hours=12)
        seconds = time_.hour * 3600 + time_.minute * 60 + time_.second
        data.extend([x for x in seconds.to_bytes(2, "big")])
        data.append(0x00)
        return bytes(data)

    def set_operating_time(self, operating_time: int) -> None:
        """set the operating time span of the station in minutes, (2-5759)"""
        self.send_command(com.SET_SYS_DATA, bytes([memaddr.ACTIVE_TIME]) + operating_time.to_bytes(2, "big"))
        self.get_system_info()

    def _splice_control_number(self, sys_data: List[int], num: int) -> List[int]:
        sys_data[memaddr.CONTROL_NUMBER] = num & 0xFF
        sys_data[memaddr.SETTINGS_FLAG] = self._setting | SETTINGS.NUMBER_HIGH_BIT if num > 255 else self._setting & ~SETTINGS.NUMBER_HIGH_BIT
        return sys_data

    def set_control_number(self, num: int) -> None:
        """set the control number of the station, supported range (1-511). IMPORTANT: some older SI Chip only support numbers till 255!"""
        if not 0 < num < 512:
            raise ValueError("Only supports values between 1 and 511")
        sys_data = self._splice_control_number(self._sys_data.copy(), num)
        self.send_command(com.SET_SYS_DATA, [int(memaddr.CONTROL_NUMBER)] + sys_data[memaddr.CONTROL_NUMBER:memaddr.SETTINGS_FLAG+1], True)
        self.get_system_info()

    def set_mode(self, mode: stationmode) -> None:
        """set the operating mode of the station"""
        sys_copy = self._sys_data.copy()
        if mode == stationmode.SIAC_TEST:
            sys_copy = self._splice_control_number(sys_copy, stationmode.SIAC_ON)
        elif is_siac_special_mode(mode):
            sys_copy = self._splice_control_number(sys_copy, mode)
            mode = stationmode.SIAC_SPECIAL
        if not is_siac_mode(mode):
            self._siac_mode = siacmode.NO_SIAC
        sys_copy[memaddr.SIAC] = misc.SIAC_ON if is_siac_mode(mode) else misc.SIAC_OFF
        sys_copy[memaddr.MODE] = mode | self.siac_mode
        self.send_command(com.SET_SYS_DATA, [int(memaddr.SIAC)] + sys_copy[memaddr.SIAC:memaddr.SETTINGS_FLAG+1], True)
        self.get_system_info()

    def set_siac_mode(self, siac_mode: siacmode) -> None:
        """set SIAC SRR behaviour (only for SIAC_CONTROL, SIAC_START, SIAC_FINISH)"""
        if not is_siac_mode(self.mode):
            raise SiError("Can't set SIAC SRR behaviour for current mode!")
        self.send_command(com.SET_SYS_DATA, [int(memaddr.MODE), self.mode | siac_mode], True)
        self.get_system_info()

    def set_protocol_setting(self, setting: CPC, set_val: bool):
        """set a single setting of the protocol settings byte"""
        sett = apply_bit_mask(self._cpc, setting, set_val << (log2(setting) - 1))
        if setting in [CPC.AUTO_SEND, CPC.HANDSHAKE]:
            setting = CPC.AUTO_SEND if setting == CPC.HANDSHAKE else CPC.HANDSHAKE
            sett = apply_bit_mask(sett, setting, int(not set_val) << (log2(setting) - 1))
        self.send_command(com.SET_SYS_DATA, [memaddr.PROTOCOL_MODE_FLAG, sett], True)

    def set_protocol_settings(self, settings: List[CPC], values: List[int]) -> None:
        """set multiple settings of the protocol settings byte at once. length of settings and values has to be equal"""
        values = [val << log2(sett - 1) for sett, val in zip(settings, values)]
        self.send_command(com.SET_SYS_DATA,
                        [memaddr.PROTOCOL_MODE_FLAG, apply_bit_mask(self._cpc, array_or([int(x) for x in settings]), array_or(values))], True)
        self.get_system_info()

    def set_setting(self, setting: SETTINGS, set_val: bool):
        """set a single setting of the settings byte"""
        self.send_command(com.SET_SYS_DATA, [memaddr.SETTINGS_FLAG, apply_bit_mask(self._setting, setting, set_val << setting)], True)

    def set_settings(self, settings: List[SETTINGS], values: List[int]) -> None:
        """set multiple settings of the settings byte at once. length of settings and values has to be equal"""
        values = [val << log2(sett - 1) for sett, val in zip(settings, values)]
        self.send_command(com.SET_SYS_DATA,
                        [memaddr.SETTINGS_FLAG, apply_bit_mask(self._setting, array_or([int(x) for x in settings]), array_or(values))], True)
        self.get_system_info()

    def turn_off(self) -> None:
        """Turn off the connected station, is only effective in remote mode (mode_direct = False)"""
        self.send_command(com.TURN_OFF, [], True)

    def trigger_feedback(self, num: int = 1) -> bool:
        """Trigger acoustic and/or visual feedback of the station <num> times"""
        if self.model_id != MODELID.SIMSSR1_AP:
            return not isinstance(self.send_command(com.FEEDBACK, [num]), NOACK)
        else:
            return True

    def ensure_readout(self) -> None:
        """ensures that the station is set to direct mode and is in mode READOUT"""
        self.flush()
        self.set_mode_direct(True)
        self.set_mode(stationmode.READOUT)
        self.set_protocol_setting(CPC.AUTO_SEND, False)
        self.get_system_info()

    def ensure_control_readout(self, control_number: int = None, enable_siac: bool = False):
        """ensures that the station is configured to receive SI Punches in Control Mode"""
        self.set_mode_direct(True)
        self.set_mode(stationmode.SIAC_CONTROL if enable_siac else stationmode.CONTROL)
        if control_number is not None:
            self.set_control_number(control_number)
        self.set_protocol_setting(CPC.AUTO_SEND, True)
        self.get_system_info()

    def wait_for_si_card(self, timeout: float = None, *, no_ack: bool = False, auto_detect_srr: bool = False) -> Optional[SICard]:
        """waits for an SiCard to be inserted, waits a maximum of timeout (infinite if unset),
        returns a SICard object or None if timeout
        if no_ack is set True, no ACK packet will be sent to the station"""
        invoke_time = time.time()
        if self._currently_waiting:
            return None
        while True:
            self._currently_waiting = True
            try:
                self.receive_command(auto_detect_srr=auto_detect_srr)
            except SiCardInserted:
                if self._si_card:
                    self._si_card.read_out_data(no_ack=no_ack)
                return self._si_card
            except SiCardRemoved:
                pass
            except SiCardAutoSend:
                return self._si_card
            finally:
                self._currently_waiting = False
            if timeout is not None and time.time() - invoke_time >= timeout:
                return None

    def sendACK(self): # pylint: disable=invalid-name
        """send a ACK packet to the station"""
        self._dev.write([proto.ACK])

    def sendNAK(self): # pylint: disable=invalid-name
        """send a NOACK packet to the station"""
        self._dev.write([proto.NAK])

    def send_command(self, command_code: com, data: Union[bytes, List[int]], retry: bool = True) -> Union[ACK, NOACK, Command]:
        """send a command to the station with data, if retry = True the command is retried a maximum of max_command_tries if receiving a NOACK response"""
        send_command(command_code, data, self._dev)
        ret = self.receive_command()
        if not retry:
            return ret
        if isinstance(ret, Command) and ret.command_code != command_code:
            return NOACK()
        if isinstance(ret, NOACK):
            self._command_tries += 1
            if self._command_tries > self.max_command_tries:
                self._command_tries = 0
                raise SiError("Max tries for command reached!")
            return self.send_command(command_code, data)
        self._command_tries = 0
        return ret

    def receive_command(self, *, auto_detect_srr: bool = False) -> Union[Command, ACK, NOACK]:
        """receive a command from the station, might raise SiCardEventError if SICard is inserted or removed while waiting on a response."""
        ret = receive_command(self._dev)
        if isinstance(ret, Command):
            if ret.command_code == com.DETECT_SI_5:
                self._si_card = SI5(ret.data[2:], self)
                raise SiCardInserted("SI-Card insert during command.")
            elif ret.command_code == com.DETECT_SI_6:
                self._si_card = SI6(ret.data[2:], self)
                raise SiCardInserted("SI-Card insert during command.")
            elif ret.command_code == com.DETECT_SI_8:
                self._si_card = SI8To10.switch_types(ret.data[2:], self)
                raise SiCardInserted("SI-Card insert during command.")
            elif ret.command_code == com.SI_REMOVED:
                raise SiCardRemoved("SI-Card removed during command.")
            elif ret.command_code == com.PUNCH_DATA:
                self._si_card = SICard.from_auto_send(ret.data, auto_detect_srr=auto_detect_srr)
                raise SiCardAutoSend("SI-Card insert with auto send enabled.")
        return ret

    def flush(self, flush_input: bool = True, flush_output: bool = True):
        """flush input and output buffer of the serial interface"""
        if flush_input:
            self._dev.flushInput()
        if flush_output:
            self._dev.flushOutput()

    @property
    def memory_size(self) -> Optional[int]:
        """system memory size of the station in bytes"""
        return getattr(self, "_mem_size", None)

    @property
    def max_command_tries(self) -> int:
        """maximum number of command retries if receiving NOACK as response"""
        return self._max_command_tries

    @max_command_tries.setter
    def max_command_tries(self, tries: int):
        self.set_max_command_tries(tries)

    def set_max_command_tries(self, tries: int):
        """set max_command_tries"""
        self._max_command_tries = tries

    @property
    def direct_mode(self) -> Optional[bool]:
        "whether the station operates in direct (True) or remote (False) mode"
        return getattr(self, "_direct_mode", None)

    @direct_mode.setter
    def direct_mode(self, set_mode_direct: bool):
        self.set_mode_direct(set_mode_direct)

    @property
    def support_extended_protocol(self) -> Optional[bool]:
        """whether the station supports the extended protocol, note that only those station are currently supported"""
        return getattr(self, "_extended_protocol", None)

    @property
    def cpc_extended_prototcol(self) -> Optional[bool]:
        """whether the extended protocol flag is set"""
        return getattr(self, "_cpc_ext_protocol", None)

    @property
    def auto_send(self) -> Optional[bool]:
        """whether autosend for readout is enabled, currently this is only supported for control read out"""
        return getattr(self, "_auto_send", None)

    @auto_send.setter
    def auto_send(self, autosend: bool) -> None:
        self.set_protocol_setting(CPC.AUTO_SEND, autosend)

    @property
    def handshake(self) -> Optional[bool]:
        """whether handshake mode for readout is enabled, currently this is the only mode supported for full SICard read out"""
        return getattr(self, "_handshake", None)

    @handshake.setter
    def handshake(self, handshake: bool):
        self.set_protocol_setting(CPC.HANDSHAKE, handshake)

    @property
    def access_with_password(self) -> Optional[bool]:
        """the the access with password flag is set"""
        return getattr(self, "_access_password", None)

    @property
    def read_after_punch(self) -> Optional[bool]:
        """whether the read after punch flag is set"""
        return getattr(self, "_read_after_punch", None)

    @read_after_punch.setter
    def read_after_punch(self, read_after_punch: bool):
        self.set_protocol_setting(CPC.READ_AFTER_PUNCH, read_after_punch)

    @property
    def optical_feedback(self) -> Optional[bool]:
        """whether optical feedback is enabled"""
        return getattr(self, "_optical_feedback", None)

    @optical_feedback.setter
    def optical_feedback(self, optical_feedback: bool):
        self.set_setting(SETTINGS.OPTICAL_FEEDBACK, optical_feedback)

    @property
    def acoustic_feedback(self) -> Optional[bool]:
        """whether acoustic feedback is enabled"""
        return getattr(self, "_acoustic_feedback", None)

    @acoustic_feedback.setter
    def acoustic_feedback(self, acoustic_feedback: bool):
        self.set_setting(SETTINGS.ACOUSTIC_FEEDBACK, acoustic_feedback)

    @property
    def operating_time(self) -> int:
        """currently configurate operating time in minutes"""
        return self._operating_time

    @operating_time.setter
    def operating_time(self, operating_time: int):
        self.set_operating_time(operating_time)

    @property
    def control_number(self) -> Optional[int]:
        """current control number"""
        return getattr(self, "_control_number", None)

    @control_number.setter
    def control_number(self, control_number: int):
        self.set_control_number(control_number)

    @property
    def mode(self) -> stationmode:
        """current operating mode"""
        return self._mode

    @mode.setter
    def mode(self, mode: stationmode) -> None:
        self.set_mode(mode)

    @property
    def siac_mode(self) -> siacmode:
        """current SIAC SRR behaviour"""
        return self._siac_mode

    @siac_mode.setter
    def siac_mode(self, siac_mode: siacmode) -> None:
        self.set_siac_mode(siac_mode)

    @property
    def model_id(self) -> MODELID:
        """bytes of model id"""
        return self._model_id

    @property
    def model_id_string(self) -> str:
        """string name of model id"""
        return self._model_id_str

    @property
    def mode_string(self) -> str:
        """string name of the current operating mode"""
        return self._mode_str

    @property
    def siac_mode_string(self) -> str:
        """string name of the current SIAC SRR behaviour"""
        return self._siac_mode_str

    @property
    def time(self) -> Optional[datetime]:
        """current time of the station, calculated with the read out time and the timedelta since the readout"""
        if self._time and self._time_read_at:
            return self._time + (datetime.now() - self._time_read_at)
        return None



"""
A library for communicating SportIdent Stations and Punch-Cards
"""

from . import comms, crc, enums, utils
from .card import *
from .station import SiUSBStation

__all__ = [
    "SiUSBStation"
]



"""
submodule for handling SI Cards and readout.
"""

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from sipyconfig.comms import Command
from sipyconfig.enums import SiCardError, SiError, com, SIAC_BEEP, siac_rev, stationmode, siacmode, mode_rev
from sipyconfig.utils import bytes_to_int

if TYPE_CHECKING:
    from sipyconfig import SiUSBStation
    from typing import Dict, List, Optional, Union, Tuple


class Punch():
    """
    Information about a control punch.
    Control Number and if available time code
    """
    control: int
    time: 'Optional[datetime]' = None
    control_mode: 'Optional[stationmode]' = None
    siac_mode: 'Optional[siacmode]' = None
    order_number: int = 0

    def __init__(
        self,
        control_number: int,
        time: 'Optional[datetime]' = None,
        control_byte: 'Optional[int]' = None, *,
        order_number: int = 0,
        auto_detect: bool = False
    ) -> None:
        self.time = time
        self.control = control_number
        self.order_number = order_number
        if control_byte is not None:
            try:
                self.control_mode = stationmode(control_byte & 0x3f)
                self.siac_mode = siacmode(control_byte & 0xf0)
            except (ValueError, TypeError):
                self.control_mode = stationmode.UNKNOWN
                self.siac_mode = siacmode.NO_SIAC
        if auto_detect and (
            self.control_mode is None or
            self.control_mode == stationmode.UNKNOWN
        ):
            if control_number == 1:
                self.control_mode = stationmode.CLEAR
            elif control_number == 2:
                self.control_mode = stationmode.CHECK
            elif control_number == 3:
                self.control_mode = stationmode.START
            elif control_number == 4:
                self.control_mode = stationmode.FINISH
        if auto_detect and self.control_mode == stationmode.SIAC_SPECIAL:
            self.control_mode = stationmode(control_number)

    @classmethod
    def from_bytes(
            cls,
            control_number: int,
            raw_time: bytes,
            punchdate: int = None,
            reftime: 'datetime' = None,
            membytes: 'Optional[bytes]' = None, *,
            auto_detect: bool = False
        ) -> 'Punch':
        """
        construct Punch class from the byte data
        """
        punchtime = SICard.decode_time(raw_time, punchdate, reftime)
        control_byte = None
        order_number = 0
        if membytes is not None:
            control_byte = membytes[0]
            order_number = membytes[1]
        return Punch(control_number, punchtime, control_byte, order_number=order_number, auto_detect=auto_detect)

    def __str__(self) -> str:
        if self.control_mode is not None:
            return f"Punch {self.control} / {mode_rev.get(self.control_mode)} at {self.time}"
        return f"Punch {self.control} at {self.time}"

    def __repr__(self) -> str:
        if self.control_mode is not None and self.siac_mode is not None:
            return f"Punch({self.control.__repr__()}, {self.time.__repr__()}, {hex(self.control_mode | self.siac_mode)})"
        return f"Punch({self.control.__repr__()}, {self.time.__repr__()})"


class SICard():
    """
    Meta class for the different SICard types
    """
    CN2: int = -1
    CN1: int = -1
    CN0: int = -1
    STD: int = -1
    ST: int = -1
    STR: int = -1
    FTD: int = -1
    FT: int = -1
    FTR: int = -1
    CTD: int = -1
    CT: int = -1
    LTD: int = -1
    LT: int = -1
    LTR: int = -1
    RC: int = -1
    P1: int = -1
    PL: int = -1
    PTR: int = -1
    PM: int = -1
    CN: int = -1
    PTD: int = -1
    PTH: int = -1
    PTL: int = -1
    BC: int = -1
    SPD: int = -1
    EPD: int = -1

    TIME_RESET = b'\xEE\xEE'

    _number: int
    _station: 'SiUSBStation'
    _start_time: 'Optional[datetime]' = None
    _start_reserve: 'Optional[datetime]' = None
    _finish_time: 'Optional[datetime]' = None
    _finish_reserve: 'Optional[datetime]' = None
    _check_time: 'Optional[datetime]' = None
    _clear_time: 'Optional[datetime]' = None
    _clear_reserve: 'Optional[datetime]' = None

    _read_at: 'Optional[datetime]' = None

    _punches: 'List[Punch]'

    _personal_data: 'Dict[str, str]'
    _other_data: 'Dict[str, Union[float, str, datetime]]'

    @staticmethod
    def decode_number(data: 'Union[bytes, List[int]]') -> int:
        """
        Handles decoding the SICard number from bytes. With SICard 5 special case
        """
        if data[0] != 0x00:
            raise SiCardError("Unknown SI-Card!")
        num = bytes_to_int(data[1:4])
        if num < 500000:
            # SI Card 5
            num2 = bytes_to_int(data[2:4])
            return data[1]*100000 + num2
        return num

    @property
    def number(self) -> int:
        """Number of the SICard"""
        return self._number

    @property
    def read_at(self) -> 'Optional[datetime]':
        """Date and time of the last read"""
        return getattr(self, "_read_at", None)

    @property
    def starttime(self) -> 'Optional[datetime]':
        """recorded start time"""
        return getattr(self, "_start_time", None)

    @property
    def starttime_reserve(self) -> 'Optional[datetime]':
        """recorded start time reserve"""
        return getattr(self, "_start_reserve", None)

    @property
    def finishtime(self) -> 'Optional[datetime]':
        """recorded finish time"""
        return getattr(self, "_finish_time", None)

    @property
    def finishtime_reserve(self) -> 'Optional[datetime]':
        """recorded finish time reserve"""
        return getattr(self, "_finish_reserve", None)

    @property
    def checktime(self) -> 'Optional[datetime]':
        """recorded check time"""
        return getattr(self, "_check_time", None)

    @property
    def cleartime(self) -> 'Optional[datetime]':
        """recorded clear time (only SI6, else see cleartime_reserve)"""
        return getattr(self, "_clear_time", None)

    @property
    def cleartime_reserve(self) -> 'Optional[datetime]':
        """recorded clear time reserve (only SI10 and up)"""
        return getattr(self, "_clear_reserve", None)

    @property
    def punches(self) -> 'List[Punch]':
        """list of recorded punches"""
        return getattr(self, "_punches", [])

    @property
    def personal_data(self) -> 'Dict[str, str]':
        """dict of the stored personal data.
        to edit data modify this dict and call write_personal_data (edit only SI10 and up)"""
        return getattr(self, "_personal_data", {})

    @property
    def other_data(self) -> 'Dict[str, str]':
        """other data stored in the SICard, such as hardware and firmware version (only SI10 and up)"""
        return getattr(self, "_other_data", {})

    def read_out_data(self, *, no_ack: bool = False) -> None:
        """read out data from SICard"""
        raise NotImplementedError

    def process_read_out(self, data: bytes, reftime: 'datetime' = None):
        """process the read out data from a SICard"""
        self._read_at = datetime.now()
        self._number = self.decode_number([0x00, data[self.CN2], data[self.CN1], data[self.CN0]])
        self._start_time = self.decode_time(data[self.ST:self.ST+2], data[self.STD] if self.STD != -1 else None, reftime)
        self._finish_time = self.decode_time(data[self.FT:self.FT+2], data[self.FTD] if self.FTD != -1 else None, reftime)
        self._check_time = self.decode_time(data[self.CT:self.CT+2], data[self.CTD] if self.CTD != -1 else None, reftime)
        if self.LT != -1:
            self._clear_time = self.decode_time(data[self.LT:self.LT+2], data[self.LTD] if self.LTD != -1 else None, reftime)
        if self.STR != -1:
            self._start_reserve = self.decode_time(data[self.STR+2:self.STR+4], data[self.STR], reftime)
        if self.FTR != -1:
            self._finish_reserve = self.decode_time(data[self.FTR+2:self.FTR+4], data[self.FTR], reftime)
        if self.LTR != -1:
            self._clear_reserve = self.decode_time(data[self.LTR+2:self.LTR+4], data[self.LTR], reftime)

        self.process_extra_data(data)

        punch_count = data[self.RC]
        if isinstance(self, SI5):
            punch_count -= 1
        if punch_count > self.PM:
            punch_count = self.PM

        self._punches = []
        if not isinstance(self, SI5):
            for punch_pointer in range(self.P1, self.P1 + data[self.RC] * self.PL, self.PL):
                punch_control = data[punch_pointer + self.CN]
                punch_time = data[punch_pointer + self.PTH:punch_pointer + self.PTL + 1]

                punch_date = None
                if self.PTD != -1:
                    punch_date = data[punch_pointer + self.PTD]
                    punch_control += (punch_date << 2) & ~0xFF

                self.add_punch(Punch.from_bytes(punch_control, punch_time, punch_date, reftime))

    def process_extra_data(self, data: bytes):
        """process any non standard data from the SICard read out"""
        raise NotImplementedError

    @classmethod
    def from_auto_send(cls, data: bytes, *, auto_detect_srr: bool = False):
        """create a SI Card object from the data send with auto send."""
        data_new = [x for x in data]
        if data_new[2] == 0x0f:
            #measures for working with SRR
            data_new[0] = data[0] & 0x7f
            data_new[2] = 0x0
        control_number = bytes_to_int(data_new[:2])
        si_number = cls.decode_number(data_new[2:6])
        time = cls.decode_time(data_new[7:9])
        mem_pointer = data[10:]
        control_byte = mem_pointer[0]
        card: SICard
        if si_number <= 499999:
            card = SI5(data_new[2:6])
        elif 500000 <= si_number <= 999999 or (2003000 <= si_number <= 2003999):
            card = SI6(data_new[2:6])
        else:
            card = SI8To10.switch_types(data_new[2:6])
        card.add_punch(Punch(control_number, time, control_byte, order_number=mem_pointer[1], auto_detect=auto_detect_srr))
        return card

    def add_punch(self, punch: Punch):
        """adds a punch to the list of punches"""
        if not self.punches:
            self._punches = []
        self._punches.append(punch)

    @classmethod
    def decode_time(cls, raw_time: 'Union[bytes, List[int]]', punchdate: int = None, reftime: 'datetime' = None) -> 'Optional[datetime]':
        """decode date and time from bytes before reftime"""
        if raw_time == cls.TIME_RESET:
            return None

        if reftime is None:
            reftime = datetime.now() + timedelta(hours=2)

        punchtime = timedelta(seconds=bytes_to_int(raw_time))

        if punchdate is not None:
            # adjust for am or pm
            if punchdate & 0b00000001:
                punchtime += timedelta(hours=12)

            # extract day of week (%7 for sunday = -1)
            day_of_week = (((punchdate & 0b00001110) >> 1) - 1) % 7

            if reftime.weekday() == day_of_week and punchtime > timedelta(hours=reftime.hour, minutes=reftime.minute, seconds=reftime.second):
                reftime -= timedelta(days=7)
            else:
                reftime -= timedelta(days=(reftime.weekday() - day_of_week) % 7)

            ref_day = reftime.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
            return ref_day + punchtime

        # no punchdate available
        ref_day = reftime.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=None)
        ref_hour = reftime - ref_day
        t_noon = timedelta(hours=12)

        if ref_hour < t_noon:
            # reference time is before noon
            if punchtime < ref_hour:
                # t is between 0:00 and t_ref
                return ref_day + punchtime
            else:
                # t is afternoon the day before
                return ref_day - t_noon + punchtime
        else:
            # reference is after noon
            if punchtime < ref_hour - t_noon:
                # t is between noon and t_ref
                return ref_day + t_noon + punchtime
            else:
                # t is in the late morning
                return ref_day + punchtime

    def __str__(self) -> str:
        return f"{self.__class__.__name__} {self.number} with {len(self.punches)} Punches"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.number} at {hex(id(self))}>"


class SI5(SICard):
    """SICard 5, numbers 4.000 - 49.999"""
    CN2 = 0x6
    CN1 = 0x4
    CN0 = 0x5
    ST = 0x13
    FT = 0x15
    CT = 0x19
    RC = 0x17
    P1 = 0x20
    PL = 0x3
    PM = 30  # punches 31-36 have no time
    CN = 0x0
    PTH = 0x1
    PTL = 0x2

    def __init__(self, data: 'Union[bytes, List[int]]', station: 'SiUSBStation' = None):
        self._number = self.decode_number(data)
        if station:
            self._station = station
        self._read_at = datetime.now()

    def read_out_data(self, *, no_ack: bool = False) -> None:
        if not self._station:
            raise AttributeError("SICard needs to be invoked with station!")

        ret = self._station.send_command(com.GET_SI_5, [], True)
        if not isinstance(ret, Command):
            raise SiError("Can't read Si-Card!")

        self.process_read_out(ret.data[2:])
        if not no_ack:
            self._station.sendACK()

    def process_read_out(self, data: bytes, reftime: 'datetime' = None):
        super().process_read_out(data, reftime)

        no_time_punches = []

        punch_count = data[self.RC] - 1
        if punch_count > self.PM:
            punch_count = self.PM
        punch_pointer = self.P1
        for _ in range(punch_count):
            if punch_pointer % 16 == 0:
                # first byte of each block is reserved for punches 31 - 36
                if data[punch_pointer]:
                    no_time_punches.append(Punch(data[punch_pointer], None))
                punch_pointer += 1

            punch_control = data[punch_pointer + self.CN]
            punch_time = data[punch_pointer + self.PTH:punch_pointer + self.PTL + 1]
            punch_date = data[punch_pointer + self.PTD] if self.PTD != -1 else None

            self.add_punch(Punch.from_bytes(punch_control, punch_time, punch_date, reftime))
            punch_pointer += self.PL
        self.punches.extend(no_time_punches)

    def process_extra_data(self, data: bytes):
        pass


class SI6(SICard):
    """SICard 6, numbers 500.000 - 999.999"""
    CN2 = 0xb
    CN1 = 0xc
    CN0 = 0xd
    STD = 0x18
    ST = 0x1a
    FTD = 0x14
    FT = 0x16
    CTD = 0x1c
    CT = 0x1e
    LTD = 0x20
    LT = 0x22
    RC = 0x12
    P1 = 0x80
    PL = 0x4
    PM = 64
    PTD = 0x0
    CN = 0x1
    PTH = 0x2
    PTL = 0x3

    _personal_structure: 'Dict[str, Tuple[int,int]]' = {
        "last_name": (0x30, 0x43),
        "first_name": (0x44, 0x57),
        "country": (0x58, 0x5B),
        "club": (0x5C, 0x7f),
        "user_id": (0x80, 0x8f),
        "phone": (0x90, 0x9F),
        "email": (0xA0, 0xC3),
        "street": (0xC4, 0xD7),
        "city": (0xD8, 0xE7),
        "zip": (0xE8, 0xEF),
        "sex": (0xF0, 0xF3),
        "birth": (0xF4, 0xFB)
    }

    SI6_STAR = b'\x01\x01\x01\x02'

    def __init__(self, data: 'Union[bytes, List[int]]', station: 'SiUSBStation' = None):
        self._number = bytes_to_int(data)
        if station:
            self._station = station
        self._read_at = datetime.now()

    def read_out_data(self, *, no_ack: bool = False) -> None:
        if not self._station:
            raise AttributeError("SICard needs to be invoked with station!")

        ret = self._station.send_command(com.GET_SI_6, [0x00], True)
        if not isinstance(ret, Command):
            raise SiError("Can't read Si-Card!")
        raw_data = ret.data[3:]

        for _ in range(2):
            ret2 = self._station.receive_command()
            if not isinstance(ret2, Command):
                raise SiError("Can't read Si-Card!")
            raw_data += ret2.data[3:]

        self.process_read_out(raw_data)
        if not no_ack:
            self._station.sendACK()

    def process_read_out(self, data: bytes, reftime: 'datetime' = None):
        super().process_read_out(data, reftime)

        self.process_extra_data(data)

        if data[:4] == self.SI6_STAR:
            raw_data = bytes([])

            for i in range(2,6):
                ret = self._station.send_command(com.GET_SI_6, [i], True)
                if not isinstance(ret, Command):
                    raise SiError("Can't read Si-Card!")
                raw_data += ret.data[3:]

            for punch_pointer in range(data[self.RC] - self.PM):
                punch_control = raw_data[punch_pointer + self.CN]
                punch_time = raw_data[punch_pointer + self.PTH:punch_pointer + self.PTL + 1]
                punch_date = raw_data[punch_pointer + self.PTD] if self.PTD != -1 else None

                self.add_punch(Punch.from_bytes(punch_control, punch_time, punch_date, reftime))
                punch_pointer += self.PL

    def process_extra_data(self, data: bytes):
        self._personal_data = {}
        for name, (from_, to_) in self._personal_structure.items():
            self._personal_data[name] = data[from_:to_+1].decode("ascii", "ignore")


class SI8To10(SICard):
    """meta class for all SICard 8 and up"""
    _data_name_list: 'Tuple[str, ...]' = ("first_name", "last_name")

    def __init__(self, data: 'Union[bytes, List[int]]', station: 'SiUSBStation' = None):
        self._number = bytes_to_int(data[1:])
        if station:
            self._station = station
        self._read_at = datetime.now()

    @classmethod
    def switch_types(cls, data: 'Union[bytes, List[int]]', station: 'SiUSBStation' = None) -> 'SI8To10':
        """switch between the SI Card Types 8 and up, via their number range"""
        num = bytes_to_int(data[1:])
        if 2000000 <= num <= 2999999:
            return SI8(data, station)
        elif 1000000 <= num <= 1999999:
            return SI9(data, station)
        elif 7000000 <= num <= 7999999:
            return SI10(data, station)
        elif 8000000 <= num <= 8999999:
            return SISIAC(data, station)
        elif 9000000 <= num <= 9999999:
            return SI11(data, station)
        else:
            raise SiCardError("Unknown SI-Card!")

    def read_out_data(self, *, no_ack: bool = False) -> None:
        if not self._station:
            raise AttributeError("SICard needs to be invoked with station!")

        raw_data = bytes([])
        for block_number in range(self.BC):
            ret = self._station.send_command(com.GET_SI_8, [block_number], True)
            if not isinstance(ret, Command):
                raise SiError("Can't read Si-Card!")
            raw_data += ret.data[3:]
            if len(raw_data) > self.P1 + self.PL * raw_data[self.RC]:
                break

        self.process_read_out(raw_data)
        if not no_ack:
            self._station.sendACK()

    def process_extra_data(self, data: bytes):
        text = data[self.SPD:self.EPD+1].decode("ascii", "ignore").split(";")

        self._personal_data = {}
        for v, k in zip(text, self._data_name_list): # pylint: disable=invalid-name
            self._personal_data[k] = v


class SI8(SI8To10):
    """SICard 8, numbers 2.000.000 - 2.999.999"""
    CN2 = 0x19
    CN1 = 0x1a
    CN0 = 0x1b
    STD = 0xc
    ST = 0xe
    FTD = 0x10
    FT = 0x12
    CTD = 0x8
    CT = 0xa
    RC = 0x16
    P1 = 0x88
    PL = 0x4
    PM = 30
    PTD = 0x0
    CN = 0x1
    PTH = 0x2
    PTL = 0x3
    BC = 2
    SPD = 0x20
    EPD = 0x87


class SI9(SI8To10):
    """SICard 9, numbers 1.000.000 - 1.999.999"""
    CN2 = 0x19
    CN1 = 0x1a
    CN0 = 0x1b
    STD = 0xc
    ST = 0xe
    FTD = 0x10
    FT = 0x12
    CTD = 0x8
    CT = 0xa
    RC = 0x16
    P1 = 0x38
    PL = 0x4
    PM = 50
    PTD = 0x0
    CN = 0x1
    PTH = 0x2
    PTL = 0x3
    BC = 2
    SPD = 0x20
    EPD = 0x37


class SI10ToSIAC(SI8To10):
    """meta class for all SICard types with the memory layout of an SICard 10 (SI10, SI11, SIAC)"""
    CN2 = 0x19
    CN1 = 0x1a
    CN0 = 0x1b
    STD = 0xc
    ST = 0xe
    STR = 0x1d8
    FTD = 0x10
    FT = 0x12
    FTR = 0x1dc
    CTD = 0x8
    CT = 0xa
    LTR = 0x1b8
    RC = 0x16
    P1 = 0x80*4
    PL = 0x4
    PM = 128
    PTD = 0x0
    CN = 0x1
    PTH = 0x2
    PTL = 0x3
    BC = 8
    SPD = 0x20
    EPD = 0x9f
    DOP_M = 0x1C
    DOP_Y = 0x1D

    _data_name_list = ("first_name", "last_name", "sex", "birth", "club", "mail", "phone", "city", "street", "zip", "country")

    def process_extra_data(self, data: bytes):
        super().process_extra_data(data)

        self._other_data = {
            "production_date": datetime(year=2000+data[self.DOP_Y], month=data[self.DOP_M], day=1)
        }

    def write_personal_data(self):
        """apply changes to personal_data dict to the SICard. Max length including separators 128"""
        if not self._station:
            raise AttributeError("SICard needs to be invoked with station!")

        text = ";".join([self._personal_data[x] for x in self._data_name_list]) + ";"
        if len(text) > 128:
            raise ValueError("Length to great maximum 128!")
        text_bytes = text.encode("ascii", "replace") + b'\xee' * (4 - len(text) % 4)
        for i in range(0, len(text), 4):
            slice_ = text_bytes[i:i+4]
            self._station.send_command(com.WRITE_SI_8, bytes([0x08 + i // 4]) + slice_)

class SI10(SI10ToSIAC):
    """SICard 10, numbers 7.000.000 - 7.999.999"""


class SI11(SI10ToSIAC):
    """SICard 11, numbers 9.000.000 - 9.999.999"""


class SISIAC(SI10ToSIAC):
    """SICard SIAC, numbers 8.000.000 - 8.999.999"""
    HW_M = 0x1C0
    HW_S = 0x1C1
    SW_M = 0x1C2
    SW_S = 0x1C3
    LOW_BAT = 0x1D5
    SIAC_MODE = 0x1CC

    def process_extra_data(self, data: bytes):
        super().process_extra_data(data)

        if not self.other_data:
            self._other_data = {}

        self._other_data["hardware_ver"] = f"{data[self.HW_M]}.{data[self.HW_S]}"
        self._other_data["software_ver"] = f"{data[self.SW_M]}.{data[self.SW_S]}"
        self._other_data["low_battery"] = data[self.LOW_BAT] == 0x6C

        if siac_rev.get(data[self.SIAC_MODE:self.SIAC_MODE+4], None):
            self._other_data["siac_mode"] = siac_rev[data[self.SIAC_MODE:self.SIAC_MODE+4]]

    def write_siac_mode(self, siac_mode: SIAC_BEEP):
        """change blinking mode of the SICard"""
        if not self._station:
            raise AttributeError("SICard needs to be invoked with station!")

        self._station.send_command(com.WRITE_SI_8, bytes([0x73]) + siac_mode.value)



"""
submodule providing enums and error classes for the whole module
"""

from typing import Dict
from enum import IntEnum, Enum


class SiError(Exception):
    """General Error for the whole module, all other Errors are derived from this class."""

class CRCError(SiError):
    """Error for not matching CRC"""

class SiCardEventError(SiError):
    """Subclass of this Error are raised, if a SI Card was inserted or removed, while waiting on a command response."""


class SiCardInserted(SiCardEventError):
    """raised when a SI Card is inserted into the station while waiting on a command response."""


class SiCardRemoved(SiCardEventError):
    """raised when the SI Card is removed from the station while waiting on a command response."""


class SiCardAutoSend(SiCardEventError):
    """raised when a SI Card is inserted in Auto Send Mode."""


class SiCardError(SiError):
    """raised when an Error occurred in parsing the SI Card information."""


class misc(IntEnum): # pylint: disable=invalid-name
    """enum for miscellaneous values"""
    MODE_DIRECT = 0x4D
    MODE_REMOTE = 0x53
    SIAC_ON = 0x07
    SIAC_OFF = 0x05

    DATA_COMMAND_OFFSET = 3


class proto(IntEnum): # pylint: disable=invalid-name
    """enum for special protocoll characters"""
    STX = 0x02  # Start of text
    ETX = 0x03  # End of text
    ACK = 0x06  # Positive Handshake return
    NAK = 0x15  # Negative Handshake return
    DLE = 0x10  # DeLimitEr for data characters
    WAKEUP = 0xFF  # send to station for wakeup


class com(IntEnum): # pylint: disable=invalid-name
    """enum for all documented command codes"""
    GET_BACK_MEM = 0x81
    SET_SYS_DATA = 0x82
    GET_SYS_DATA = 0x83
    SRR_WRITE = 0xA2
    SRR_READ = 0xA3
    SRR_QUERY = 0xA6
    SRR_PING = 0xA7
    SRR_ADHOC = 0xA8
    GET_SI_5 = 0xB1
    WRITE_SI_5 = 0xC3
    PUNCH_DATA = 0xD3
    CLEAR_SI = 0xE0
    GET_SI_6 = 0XE1
    DETECT_SI_5 = 0xE5
    DETECT_SI_6 = 0xE6
    SI_REMOVED = 0xE7
    DETECT_SI_8 = 0xE8
    WRITE_SI_8 = 0xEA
    GET_SI_8 = 0xEF
    SET_MS_MODE = 0XF0
    GET_MS_MODE = 0xF1
    ERASE_BACK = 0xF5
    SET_TIME = 0xF6
    GET_TIME = 0xF7
    TURN_OFF = 0xF8
    FEEDBACK = 0xF9
    SET_BAUDRATE = 0xFE
    # Basic Commands
    SET_MS_MODE_BASIC = 0x70
    SET_BAUDRATE_BASIC = 0x7E


com_rev: Dict[int, str] = {v.value: k for k, v in com.__members__.items()}


class memaddr(IntEnum): # pylint: disable=invalid-name
    """enum for all documented system data memory addresses in SI Stations"""
    SERIAL_NO = 0x00
    SRR_CFG = 0x04
    FIRMWARE = 0x05
    BUILD_DATE = 0x08
    MODEL_ID = 0x0B
    MEM_SIZE = 0x0D
    BAT_DATE = 0x15
    BAT_CAP = 0x19
    BACKUP_PTR_HI = 0x1C
    BACKUP_PTR_LO = 0x21
    SI6_CB = 0x33
    SRR_CHANNEL = 0x34
    USED_BAT_CAP = 0x35
    MEM_OVERFLOW = 0x3D
    SRR_PROTOCOL = 0x3F
    BAT_VOLT = 0x50
    SIAC = 0x5A
    PROGRAM = 0x70
    MODE = 0x71
    CONTROL_NUMBER = 0x72
    SETTINGS_FLAG = 0x73
    PROTOCOL_MODE_FLAG = 0x74
    WAKEUP_DATE = 0x75
    WAKEUP_TIME = 0x78
    SLEEP_TIME = 0x7B
    ACTIVE_TIME = 0x7E


class memlen(IntEnum): # pylint: disable=invalid-name
    """enum for the length of the values associated with the address described in memaddr"""
    SERIAL_NO = 0x04
    SRR_CFG = 0x01
    FIRMWARE = 0x03
    BUILD_DATE = 0x03
    MODEL_ID = 0x02
    MEM_SIZE = 0x01
    BAT_DATE = 0x03
    BAT_CAP = 0x02
    BACKUP_PTR_HI = 0x02
    BACKUP_PTR_LO = 0x02
    SI6_CB = 0x01
    SRR_CHANNEL = 0x01
    USED_BAT_CAP = 0x03
    MEM_OVERFLOW = 0x01
    SRR_PROTOCOL = 0x01
    BAT_VOLT = 0x02
    SIAC = 0x01
    PROGRAM = 0x01
    MODE = 0x01
    CONTROL_NUMBER = 0x01
    SETTINGS_FLAG = 0x01
    PROTOCOL_MODE_FLAG = 0x01
    WAKEUP_DATE = 0x03
    WAKEUP_TIME = 0x03
    SLEEP_TIME = 0x03
    ACTIVE_TIME = 0x02


class stationmode(IntEnum): # pylint: disable=invalid-name
    """enum for all supported operating modes of the SI station"""
    CONTROL = 0x02
    START = 0x03
    FINISH = 0x04
    READOUT = 0x05
    CLEAR_OLD = 0x06
    CLEAR = 0x07
    CHECK = 0x0A
    PRINTOUT = 0x0B
    START_TRIG = 0x0C
    FINISH_TRIG = 0x0D
    SIAC_TEST = 0x11
    SIAC_CONTROL = 0x32
    SIAC_START = 0x33
    SIAC_FINISH = 0x34
    SIAC_SPECIAL = 0x01
    SIAC_BATTERY_TEST = 0x7B
    SIAC_ON = 0x7C
    SIAC_OFF = 0x7D
    SIAC_RADIO_READOUT = 0x7F
    UNKNOWN = 0xFF


mode_rev: Dict[int, str] = {
    v.value: k for k, v in stationmode.__members__.items()}


class siacmode(IntEnum): # pylint: disable=invalid-name
    """enum for different SIAC radio modes"""
    NO_SIAC=0x00
    SIAC_SPECIAL=0x10
    NO_RADIO=0x30
    LAST_RECORD=0x70
    ALL_RECORDS=0xb0
    UNSENT_RECORDS=0xf0

siacmode_rev: Dict[int, str] = {
    v.value: k for k, v in siacmode.__members__.items()}


class MODELID(Enum): # pylint: disable=invalid-name
    """enum for station model id"""
    SIMSSR1_AP = b'\x6F\x21'
    BSF3 = b'\x80\x03'
    BSF4 = b'\x80\x04'
    BSM4_RS232 = b'\x80\x84'
    BSM6_USB = b'\x80\x86'
    BSF5 = b'\x81\x15'
    BSF7 = b'\x81\x17'
    BSF8 = b'\x81\x18'
    BSF6 = b'\x81\x46'
    BSF7_SI_MASTER = b'\x81\x87'
    BSF8_SI_MASTER = b'\x81\x88'
    BSF7_2 = b'\x81\x97'
    BSF8_2 = b'\x81\x98'
    BSM7_USB = b'\x91\x97'
    BSM8_USB = b'\x91\x98'
    BS7_S = b'\x95\x97'
    BS11_BL = b'\x9D\x9A'
    BS7_P = b'\xB1\x97'
    BS8_P = b'\xB1\x98'
    BS7_GSM = b'\xB8\x97'
    BS11_BS_RED = b'\xCD\x9B'
    UNKNOWN = b'\xff\xff'


model_rev: Dict[bytes, str] = {
    v.value: k for k, v in MODELID.__members__.items()}


class CPC(IntEnum):
    """enum of the bit mask for the communications protocol byte"""
    EXTENDED_PROTCOL = 0x01
    AUTO_SEND = 0x02
    HANDSHAKE = 0x04
    ACCESS_PASSWORD = 0x10
    READ_AFTER_PUNCH = 0x40


class SETTINGS(IntEnum):
    """enum of the bit mask for the settings byte"""
    OPTICAL_FEEDBACK = 0x01
    ACOUSTIC_FEEDBACK = 0x04
    NUMBER_HIGH_BIT = 0x40


class coderange(IntEnum): # pylint: disable=invalid-name
    """suggested code range for different operating modes"""
    READOUT = 15
    PRINTOUT = 11
    CLEAR = 1
    START_FINISH_CHECK_FROM = 2
    START_FINISH_CHECK_TO = 30
    CONTROL_FROM = 31
    CONTROL_TO = 511


class defaulttime(IntEnum): # pylint: disable=invalid-name
    """"enum for the default/suggested operating times for the operating modes"""
    CLEAR = 4
    CHECK = 4
    START = 4
    CONTROL = 4
    FINISH = 4
    READOUT = 1
    BEACON_START = 12
    BEACON_CONTROL = 12
    BEACON_FINISH = 12
    SIAC_BATTERY = 1
    SIAC_ON = 1
    SIAC_OFF = 1
    SIAC_READOUT = 1
    SIAC_TEST = 12


class SIAC_BEEP(Enum): # pylint: disable=invalid-name
    """enum for the different SIAC feedback (beep) modes"""
    DEFAULT = b'\x18\x00\x05\x4A'
    SHORT = b'\x06\x00\x05\x4A'
    LONG = b'\x30\x00\x05\x4A'
    BLINK_ONLY = b'\x20\x00\x66\x00'

siac_rev: Dict[bytes, str] = {v.value: k for k, v in SIAC_BEEP.__members__.items()}



# pylint: disable=C0114,C0115,C0116,invalid-name,too-few-public-methods,no-name-in-module,attribute-defined-outside-init

from typing import Union, List
from serial import SerialException, SerialTimeoutException

from PyQt5.QtWidgets import (QApplication, QWidget, QMainWindow, QLabel, QScrollArea,
                            QCheckBox, QVBoxLayout, QHBoxLayout, QGridLayout, QTabWidget,
                            QSpinBox, QRadioButton, QPushButton, QButtonGroup, QComboBox,
                            QLineEdit, QErrorMessage)
from PyQt5.QtCore import QThread, pyqtSignal, QTimer, Qt
from PyQt5.QtGui import QMouseEvent, QIcon

from sipyconfig import SiUSBStation
from sipyconfig.card import SICard
from sipyconfig.comms import ACK, NOACK, Command
from sipyconfig.station import SiError
from sipyconfig.enums import MODELID, siacmode, stationmode, com, com_rev
from sipyconfig.utils import is_siac_mode, is_siac_special_mode, evenhex

class ReadOutThread(QThread):
    cardRead = pyqtSignal(SICard)
    station: SiUSBStation
    should_stop: bool = False

    def setStation(self, station: SiUSBStation):
        self.station = station

    def run(self):
        self.station.ensure_readout()
        while not self.should_stop:
            try:
                ret = self.station.wait_for_si_card(0.1)
                if ret:
                    self.cardRead.emit(ret)
            except SiError:
                pass
        self.should_stop = False

    def setExit(self, do_exit: bool):
        self.should_stop = do_exit

class ReadOutWidget(QScrollArea):
    lineSelected = pyqtSignal(int)
    cardSelected = pyqtSignal(SICard)

    _cards: List[SICard]

    def __init__(self,) -> None:
        super().__init__()
        self._cards = []

        self.main_lay = QGridLayout()

        self._addHeader()

        wid = QWidget()
        wid.setLayout(self.main_lay)
        self.setWidget(wid)

        self.setWidgetResizable(True)

    def _addHeader(self):
        for i, wid in enumerate([QLabel(x) for x in ["#", "Read At", "SI-ID", "First Name", "Last Name", "Records", "Voltage"]]):
            self.main_lay.addWidget(wid, 0, i)
        self.main_lay.setRowStretch(1, 1)
        self.main_lay.setColumnStretch(1, 1)
        self.main_lay.setColumnStretch(3, 1)
        self.main_lay.setColumnStretch(4, 1)

    def _addLine(self, data: 'list'):
        line = self.main_lay.rowCount() # 2
        self.main_lay.setRowStretch(line - 1, 0)

        def mouseEvent(ev: QMouseEvent, line: int = line) -> None:
            if ev.button() == QMouseEvent.LeftButton: # type: ignore
                self.lineSelected.emit(line - 1)

        widgets = [QLabel(str(line - 1))] + [QLabel(str(dat)) for dat in data]
        for i, wid in enumerate(widgets):
            wid.mousePressEvent = mouseEvent # type: ignore
            self.main_lay.addWidget(wid, line - 1, i)
        self.main_lay.setRowStretch(line, 1)

    def addSiCard(self, card: 'SICard'):
        data = [card.read_at, card.number, "", "", len(card.punches), ""]
        self._addLine(data)

class MainWindow(QMainWindow):

    def __init__(self, device: str = None) -> None:
        super().__init__()

        self.setWindowTitle("SPORTident Py-Config")
        try:
            self.setWindowIcon(QIcon("sportident_python.png"))
        except FileNotFoundError:
            pass

        self.initUI()
        self.initSerial(device)

    def initUI(self):
        self.main_wid = QTabWidget()

        self.initControlWid()
        self.initSiReadOut()
        self.initDebugWid()

        self.setCentralWidget(self.main_wid)

    def initControlWid(self):
        self.control_lay = QGridLayout()

        select = QButtonGroup()
        self.direct_select = QRadioButton("Direct")
        select.addButton(self.direct_select)
        self.remote_select = QRadioButton("Remote")
        select.addButton(self.remote_select)
        self.direct_select.click()
        self.direct_select.toggled.connect(self.switchRemote)
        self.control_lay.addWidget(self.direct_select, 0, 0)
        self.control_lay.addWidget(self.remote_select, 0, 1)

        reread = QPushButton("read again")
        reread.pressed.connect(self.updateGui)
        self.control_lay.addWidget(reread, 0, 2)

        self.time_wid = QLabel()
        self.control_lay.addWidget(self.time_wid, 1, 0, 1, 2)
        self.time_set_wid = QPushButton("Set Time")
        self.time_set_wid.pressed.connect(self.setTime)
        self.control_lay.addWidget(self.time_set_wid, 1, 2)

        self.time_timer = QTimer(self)
        self.time_timer.timeout.connect(self.updateTime)
        self.time_timer.start(500)

        self.control_lay.addWidget(QLabel("Control Number:"), 2, 0)
        self.control_num_wid = QSpinBox()
        self.control_num_wid.setMinimum(1)
        self.control_num_wid.setMaximum(511)
        self.control_lay.addWidget(self.control_num_wid, 2, 1)

        self.station_mode_wid = QComboBox()
        self.station_mode_wid.addItems(stationmode.__members__.keys())
        self.station_mode_wid.currentTextChanged.connect(self.updateGuiStationMode)
        self.control_lay.addWidget(self.station_mode_wid, 3, 0, 1, 2)

        self.turn_off_wid = QPushButton("Turn Off")
        self.turn_off_wid.pressed.connect(self.turnOff)
        self.control_lay.addWidget(self.turn_off_wid, 3, 2)

        self.siac_mode_wid = QComboBox()
        self.siac_mode_wid.addItems(siacmode.__members__.keys())
        self.control_lay.addWidget(self.siac_mode_wid, 4, 0, 1, 2)

        apply = QPushButton("Apply changes")
        apply.pressed.connect(self.applyChanges)
        self.control_lay.addWidget(apply, 5, 0, 1, 2)

        self.control_lay.setRowStretch(self.control_lay.rowCount(), 1)
        self.control_lay.setColumnStretch(self.control_lay.columnCount(), 1)
        self.control_wid = QWidget()
        self.control_wid.setLayout(self.control_lay)

        self.main_wid.addTab(self.control_wid, "SI-Station")

    def initSiReadOut(self):
        self.read_out_lay = QVBoxLayout()

        self.read_out_wid = QWidget()
        self.read_out_wid.setLayout(self.read_out_lay)

        self.read_out_enable_wid = QCheckBox("Enable Read Out")
        self.read_out_enable_wid.toggled.connect(self.toggleReadOut)
        self.read_out_lay.addWidget(self.read_out_enable_wid)

        self.read_out_sub_lay = QHBoxLayout()
        self.read_out_table = ReadOutWidget()
        self.read_out_details = QLabel()
        self.read_out_sub_lay.addWidget(self.read_out_table)
        self.read_out_sub_lay.addWidget(self.read_out_details)
        self.read_out_details.hide()
        self.read_out_lay.addLayout(self.read_out_sub_lay)

        self.read_out_thread = ReadOutThread()
        self.read_out_thread.cardRead.connect(self.read_out_table.addSiCard)

        self.main_wid.currentChanged.connect(lambda x: self.toggleReadOut(False) if x != 1 else '')

        self.main_wid.addTab(self.read_out_wid, "Read Out")

    def initDebugWid(self):
        self.debug_lay = QGridLayout()

        self.debug_command_wid = QComboBox()
        self.debug_command_wid.addItems(com.__members__.keys())
        self.debug_lay.addWidget(self.debug_command_wid, 0, 0)

        self.debug_lay.addWidget(QLabel("Value"), 1, 0)
        self.debug_value_wid = QLineEdit()
        self.debug_lay.addWidget(self.debug_value_wid, 2, 0)

        self.debug_apply_wid = QPushButton("Write!")
        self.debug_apply_wid.pressed.connect(self.debugApply)
        self.debug_lay.addWidget(self.debug_apply_wid, 3, 0)

        self.debug_read_output = QLabel()
        self.debug_read_output.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.debug_lay.addWidget(self.debug_read_output, 4, 0)
        self.debug_read_wid = QPushButton("Read!")
        self.debug_read_wid.pressed.connect(self.debugRead)
        self.debug_lay.addWidget(self.debug_read_wid, 5, 0)

        self.debug_lay.setRowStretch(self.debug_lay.rowCount(), 1)
        self.debug_wid = QWidget()
        self.debug_wid.setLayout(self.debug_lay)

        self.main_wid.addTab(self.debug_wid, "Debug")

    def initSerial(self, device: str = None):
        self.si = SiUSBStation(device)
        self.si.max_command_tries = 10

        self.read_out_thread.setStation(self.si)

        self.updateGui()

    def updateGui(self):
        try:
            self.si.get_system_info()
            self.si.trigger_feedback(1)
        except SiError as e:
            self.showError(e)
            return
        except SerialException:
            self.initSerial()

        self.control_num_wid.setValue(self.si.control_number)
        self.control_num_wid.setDisabled(self.si.mode == stationmode.SIAC_SPECIAL or self.si.mode == stationmode.SIAC_TEST)
        self.station_mode_wid.setCurrentText(self.si.mode_string)
        self.siac_mode_wid.setCurrentText(self.si.siac_mode_string)
        self.turn_off_wid.setDisabled(bool(self.si.direct_mode))

        self.control_num_wid.setDisabled(self.si.model_id == MODELID.SIMSSR1_AP)
        self.direct_select.setDisabled(self.si.model_id == MODELID.SIMSSR1_AP)
        self.remote_select.setDisabled(self.si.model_id == MODELID.SIMSSR1_AP)
        self.station_mode_wid.setDisabled(self.si.model_id == MODELID.SIMSSR1_AP)

    def updateGuiStationMode(self, mode: str):
        self.control_num_wid.setDisabled(is_siac_special_mode(getattr(stationmode, mode)))
        if is_siac_mode(getattr(stationmode, mode)) and not is_siac_special_mode(getattr(stationmode, mode)):
            if not self.siac_mode_wid.isEnabled():
                self.siac_mode_wid.setDisabled(False)
                self.siac_mode_wid.setCurrentText("NO_RADIO")
        else:
            self.siac_mode_wid.setCurrentText("NO_SIAC")
            self.siac_mode_wid.setDisabled(True)

    def switchRemote(self, direct: bool):
        try:
            self.si.set_mode_direct(direct)
            self.updateGui()
        except SiError as e:
            self.showError(e)
        except SerialException:
            self.initSerial()

    def turnOff(self):
        try:
            self.si.turn_off()
        except SiError as e:
            self.showError(e)

    def setTime(self):
        try:
            self.si.set_time()
            self.updateGui()
        except SiError as e:
            self.showError(e)

    def updateTime(self):
        new_time = self.si.time
        if new_time:
            new_time.replace(microsecond=0)
            self.time_wid.setText(str(new_time))

    def toggleReadOut(self, enable: bool):
        try:
            if enable:
                self.read_out_thread.setExit(False)
                self.read_out_thread.start()
            else:
                if self.read_out_enable_wid.isChecked():
                    self.read_out_enable_wid.setChecked(False)
                self.read_out_thread.setExit(True)
        except SiError as e:
            self.showError(e)

    def showError(self, ex: Exception):
        dia = QErrorMessage(self)
        dia.showMessage(str(ex))

    def applyChanges(self):
        try:
            self.si._siac_mode = getattr(siacmode, self.siac_mode_wid.currentText()) # pylint: disable=protected-access
            self.si.mode = getattr(stationmode, self.station_mode_wid.currentText())
            if self.si.mode != stationmode.SIAC_SPECIAL and self.si.mode != stationmode.SIAC_TEST:
                self.si.control_number = self.control_num_wid.value()
        except SiError as e:
            self.showError(e)
        self.updateGui()

    def debugApply(self):
        command = getattr(com, self.debug_command_wid.currentText())
        a = self.debug_value_wid.text()
        if len(a) % 2 != 0:
            self.showError(ValueError("Need even number of characters to convert to bytes!"))
            return
        value = [int(a[x:x+2], 16) for x in range(0,len(a),2)]
        try:
            self.writeDebugRead(self.si.send_command(command, value, True))
        except SiError as e:
            self.showError(e)
        self.updateGui()

    def writeDebugRead(self, ret: Union[ACK, NOACK, Command]):
        if isinstance(ret, ACK):
            self.debug_read_output.setText("ACK")
        elif isinstance(ret, NOACK):
            self.debug_read_output.setText("NOACK")
        else:
            self.debug_read_output.setText(com_rev.get(ret.command_code, evenhex(ret.command_code)) + " - " + " ".join(evenhex(x) for x in ret.data))

    def debugRead(self):
        try:
            self.writeDebugRead(self.si.receive_command())
        except (SiError, SerialException, SerialTimeoutException) as e:
            self.showError(e)


if __name__ == "__main__":
    import sys
    from os import environ

    environ["QT_DEVICE_PIXEL_RATIO"] = "0"
    environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
    environ["QT_SCREEN_SCALE_FACTORS"] = "1"
    environ["QT_SCALE_FACTOR"] = "1"

    app = QApplication([])

    window = MainWindow(sys.argv[1] if len(sys.argv) > 1 else None)
    window.show()

    sys.exit(app.exec_())



"""
utility functions for the library
"""

from typing import TYPE_CHECKING

from sipyconfig.enums import stationmode

if TYPE_CHECKING:
    from sipyconfig.card import SICard
    from typing import Iterable, List, Union

def log2(number: int) -> int:
    """returns length of number in binary without leading zeros"""
    return len(format(number, "b"))

def apply_bit_mask(number: int, mask: int, value: int) -> int:
    """sets number to value if mask (works bitwise)"""
    length = max(log2(x) for x in [number, mask, value])
    num_bin = format(number, f"0{length}b")
    mask_bin = format(mask, f"0{length}b")
    value_bin = format(value, f"0{length}b")
    result = "".join(v if int(m) else n for n, m,
                     v in zip(num_bin, mask_bin, value_bin))
    return int(result, 2)

def array_or(array: 'List[int]') -> int:
    """return bitwise or of all elements of the list"""
    result = 0
    for num in array:
        result |= num
    return result

def is_siac_special_mode(station_mode: int) -> bool:
    """returns True if mode is a SIAC special mode"""
    return station_mode in [stationmode.SIAC_BATTERY_TEST, stationmode.SIAC_ON, stationmode.SIAC_OFF, stationmode.SIAC_RADIO_READOUT, stationmode.SIAC_TEST]

def is_siac_mode(station_mode: int) -> bool:
    """returns True if mode is a SIAC mode"""
    return is_siac_special_mode(station_mode) or station_mode in [stationmode.SIAC_CONTROL, stationmode.SIAC_START,
                                                                stationmode.SIAC_FINISH, stationmode.SIAC_TEST, stationmode.SIAC_SPECIAL]

def evenhex(num: int) -> str:
    """return hex() of num with leading zeros to make len() even"""
    hex_text = hex(num)[2:]
    if len(hex_text) % 2 != 0:
        return f"0x0{hex_text}"
    return f"0x{hex_text}"

def hextoascii(num: int) -> str:
    """return ascii of num if num between 32 and 126 ('normal' character)"""
    if 32 <= num <= 126:
        return chr(num)
    return evenhex(num)

def bytes_to_int(data: 'Union[bytes, List[int]]') -> int:
    """return int of bytes"""
    value = 0
    for offset, byte in enumerate([x for x in data][::-1]):
        value += byte << offset*8
    return value

def print_si_card(card: 'SICard'):
    """pretty print data stored in an SICard Object"""
    print(f"""
{card}
clear:  {card.cleartime}
check:  {card.checktime}
start:  {card.starttime}
finish: {card.finishtime}

clear_res:  {card.cleartime_reserve}
start_res:  {card.starttime_reserve}
finish_res: {card.finishtime_reserve}

personal data: {card.personal_data}
other data:    {card.other_data}

{len(card.punches)} punches:""")

    for i, punch in enumerate(card.punches):
        print(f"{i+1:03d} - {punch}")

def combine_punches(cards: 'Iterable[SICard]') -> 'List[SICard]':
    """combines all punches of the same card"""
    ret: 'List[SICard]' = []
    for card in cards:
        if len(ret) == 0:
            ret.append(card)
            continue
        if ret[-1].number != card.number:
            ret.append(card)
            continue
        ret[-1].punches.extend(card.punches)
        ret[-1].punches.sort(key=lambda a: a.order_number)
    return ret



"""
submodule for the serial communication with the SI Station
"""

from typing import Union, List
from serial import Serial, SerialTimeoutException

from sipyconfig.crc import compute_crc
from sipyconfig.enums import com, proto, CRCError

class ACK():
    """class representing a ACK response"""


class NOACK():
    """class representing a NOACK response"""


class Command():
    """class representing a received Command or a Command about to be sent"""
    START_SEQUENCE = [proto.WAKEUP, proto.STX, proto.STX]

    _command_code: int
    _data: List[int]
    _crc: bytes
    _is_extended_command: bool
    _is_receiving_command: bool

    def __init__(self, command_code: int, data: Union[bytes, List[int]] = None, *, receiving: bool = False) -> None:
        self._is_receiving_command = receiving
        self.command_code = command_code
        if data is not None:
            if isinstance(data, bytes):
                self._data = [x for x in data]
            else:
                self._data = data
            if not self._is_receiving_command:
                self.compute_crc()

    @property
    def command_code(self) -> int:
        """command code of the command"""
        return self._command_code

    @command_code.setter
    def command_code(self, command_code: int) -> None:
        self._command_code = command_code
        self._is_extended_command = not self.is_basic_protocol(command_code)

    @property
    def is_extended_command(self) -> bool:
        """whether this command is part of the extended protocol"""
        return self._is_extended_command

    @property
    def data(self) -> bytes:
        """data of the command"""
        return bytes(self._data)

    @data.setter
    def data(self, data: List[int]) -> None:
        if len(data) > 255:
            raise ValueError("Datastream to big, can only send up to 255 bytes at a time.")
        self._data = data
        if not self._is_receiving_command:
            self.compute_crc()

    def add_data(self, data: Union[int, List[int]]) -> None:
        """append data to this command"""
        if not getattr(self, "_data", None):
            self._data = []
        if isinstance(data, int):
            if self.length + 1 > 255:
                raise ValueError("Datastream to big, can only send up to 255 bytes at a time.")
            self._data.append(data)
        else:
            if self.length + len(data) > 255:
                raise ValueError("Datastream to big, can only send up to 255 bytes at a time.")
            self._data.extend(data)
        if not self._is_receiving_command:
            self.compute_crc()

    def compute_crc(self) -> bytes:
        """compute the crc of the currently stored data"""
        self._crc = compute_crc(self.content).to_bytes(2, "big")
        return self.crc

    @property
    def content(self) -> bytes:
        """return [command_code, length, <data>] as bytes"""
        return bytes([self.command_code, self.length]) + self.data

    @property
    def crc(self) -> bytes:
        """currently stored crc"""
        return self._crc

    @crc.setter
    def crc(self, crc: bytes) -> None:
        if not self._is_receiving_command:
            raise TypeError("Setting CRC is only possible when command is set receiving!")
        self._crc = crc

    @property
    def length(self) -> int:
        """length of the stored data"""
        return len(self._data)

    @property
    def packet(self) -> 'bytes':
        """the full byte sequence to be sent to the station, (content + crc / data with DLE)"""
        if self.is_extended_command:
            packet = bytes([self.command_code, self.length]) + bytes(self.data) + self.crc
        else:
            packet = bytes([self.command_code]) + bytes(sum([[proto.DLE, x] for x in self.data], []))
        return bytes(self.START_SEQUENCE) + packet + bytes([proto.ETX])

    def check_crc(self, crc: 'Union[int, bytes]' = None, content: bytes = None) -> bool:
        """check crc of content, omitted parameters are taken from stored values"""
        if content is None:
            content = self.content
        if crc is None:
            crc = compute_crc(content).to_bytes(2, "big")
        if isinstance(crc, int):
            crc = crc.to_bytes(2, "big")
        return self.crc == crc

    @staticmethod
    def is_basic_protocol(command_code: int) -> bool:
        '''command_code < 0x80 or 0xC4'''
        return command_code < 0x80 or command_code == 0xC4

    @staticmethod
    def is_extended_protocol(command_code: int) -> bool:
        '''command_code >= 0x80 and not 0xC4'''
        return command_code >= 0x80 and not command_code == 0xC4

    @staticmethod
    def strip_data_of_dle(data: Union[bytes, List[int]]) -> List[int]:
        """remove DLE from a data stream so it can be processed further"""
        ret = []
        seen_dle = False
        for dat in data:
            if not seen_dle and dat == proto.DLE:
                seen_dle = True
                continue
            ret.append(dat)
            seen_dle = False
        return ret



def receive_basic_command(command_code: int, dev: 'Serial') -> 'Command':
    '''receive command part of the basic protocol,
    structure: STX | command code | parameter/data with DLE | ETX'''
    command = Command(command_code, receiving=True)
    while True:
        byte = dev.read(1)[0]
        if byte == proto.DLE:
            command.add_data(dev.read(1))
        elif byte == proto.ETX:
            return command


def receive_extended_command(command_code: int, dev: 'Serial') -> 'Command':
    '''receive command part of the extended protocol,
    structure: STX | command code | length byte | parameter/data | CRC1 | CRC0 | ETX'''
    command = Command(command_code, receiving=True)
    length = dev.read(1)[0]
    command.data = dev.read(length)
    command.crc= dev.read(2)
    if not command.check_crc():
        raise CRCError("CRC does not match!")
    while dev.read(1)[0] != proto.ETX:
        pass
    return command


def receive_command(dev: 'Serial') -> 'Union[Command, ACK, NOACK]':
    """receive a command, switches internally between receive_basic_command and receive_extended_command, depending on the command code"""
    try:
        have_seen_stx = False
        while True:
            byte = dev.read(1)
            if len(byte) == 0:
                raise SerialTimeoutException
            byte = byte[0]
            if byte == proto.STX:
                have_seen_stx = True
            elif have_seen_stx:
                if Command.is_basic_protocol(byte):
                    return receive_basic_command(byte, dev)
                elif Command.is_extended_protocol(byte):
                    return receive_extended_command(byte, dev)
            elif not have_seen_stx and byte == proto.NAK:
                return NOACK()
    except SerialTimeoutException:
        return NOACK()


def send_command(command_code: com, data: Union[bytes, List[int]], dev: Serial) -> int:
    '''sends given data to the station via the extended or basic protocoll, depending on the command code'''
    command = Command(command_code, data)
    return dev.write(command.packet)



"""
setup.py for python module
"""
import setuptools

with open("README.md", encoding="UTF-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sipyconfig",
    version="1.1.0",
    author="Teichi",
    author_email="tobias@teichmann.top",
    description="Python replacement for the functionality of SportIdent Config+",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Teigi/sipyconfig",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 1 - Planning",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
        "Typing :: Typed",
        "Topic :: Utilities"
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=["pyserial"]
)



"""
CRC module for communication with SIStations
"""

from typing import Union, List

def compute_crc(data: Union[bytes, List[int]]) -> 'int':
    '''compute the CRC by Spec'''
    if isinstance(data, bytes):
        data = [x for x in data]
    i_tmp, ui_tmp1, ui_val = 0, 0, 0
    i = 0
    ui_count = len(data)

    if ui_count < 2:
        return 0

    ui_tmp1 = data[i]
    i += 1
    ui_tmp1 = (ui_tmp1 << 8) + data[i]
    i += 1

    if ui_count == 2:
        return ui_tmp1

    for i_tmp in range(ui_count >> 1, 0, -1):
        if i_tmp > 1:
            ui_val = data[i]
            i += 1
            ui_val = (ui_val << 8) + data[i]
            i += 1
        else:
            if ui_count & 1:
                ui_val = data[i]
                ui_val <<= 8
            else:
                ui_val = 0

        for _ in range(16):
            if ui_tmp1 & 0x8000:
                ui_tmp1 <<= 1
                ui_tmp1 &= 0x0000FFFF
                if ui_val & 0x8000:
                    ui_tmp1 += 1
                    ui_tmp1 &= 0x0000FFFF
                ui_tmp1 ^= 0x8005
            else:
                ui_tmp1 <<= 1
                ui_tmp1 &= 0x0000FFFF
                if ui_val & 0x8000:
                    ui_tmp1 += 1
                    ui_tmp1 &= 0x0000FFFF
            ui_val <<= 1
            ui_val &= 0x0000FFFF
    return ui_tmp1
